import numpy as np
import pandas as pd
from gmat.gmatrix import agmat
from gmat.mvlmm.unbalance import fixed
import logging
logging.basicConfig(level=logging.INFO)

bed_file = '/nfs1/ningc/WORK/Acode/GMAT/PKG/examples/data/yeast/CobaltChloride'


data_file = '/nfs1/ningc/WORK/Acode/GMAT/PKG/examples/data/yeast/CobaltChloride12'
id = 'ID'
trait = ['trait1', 'trait2']
agmat_file = '/nfs1/ningc/WORK/Acode/GMAT/PKG/examples/data/yeast/test'
var_com = pd.read_csv("/nfs1/ningc/WORK/Acode/GMAT/PKG/examples/data/yeast/mvlmm_varcom.var", sep='\s+', header=0)
prefix_outfile = '/nfs1/ningc/WORK/Acode/GMAT/PKG/examples/data/yeast/fixed3.res'
fixed(data_file, id, trait, bed_file, agmat_file, var_com, snp_lst=None, fix=None, prefix_outfile=prefix_outfile)



import logging
from gmat.gmatrix import agmat
from gmat.uvlmm.gwas import fixed, fixed_gbye
from gmat.uvlmm.varcom import weight_emai_eigen


logging.basicConfig(level=logging.INFO)
pheno_file = 'pheno'
bed_file = 'plink'


agmat_file = 'test'
agmat2 = agmat(bed_file, out_file=agmat_file, inv=False, small_val=0.001, out_fmt='id_id_val')

weight_emai_eigen(pheno_file, agmat_file, class_lst=["mean", "sex", "treat"], \
fix="mean + sex + age + treat", maxiter0=100)



import numpy as np

matrix = np.array(
    [
        [0, 1, 1, 0],
        [0, 0, 1, 0],
        [2, 1, 1, 1],
    ])


for i in range(matrix.shape[1]):
    for j in range(matrix.shape[1]):
        if i != j:
            inner_product = np.inner(matrix[:, i], matrix[:, j])
            norm_i = np.linalg.norm(matrix[:, i])
            norm_j = np.linalg.norm(matrix[:, j])
            if np.abs(inner_product - norm_j * norm_i) < 1E-5:
                print(i, j, 'Dependent')
            else:
                print(i, j, 'Independent')





import numpy as np
import pandas as pd
from gmat.gmatrix import agmat
from gmat.mvlmm.unbalance import varcom
import logging
logging.basicConfig(level=logging.INFO)

bed_file = '/nfs1/ningc/WORK/Acode/GMAT/PKG/examples/data/yeast/CobaltChloride'
agmat(bed_file, out_file='/nfs1/ningc/WORK/Acode/GMAT/PKG/examples/data/yeast/test', out_fmt='id_id_val')

data_file = '/nfs1/ningc/WORK/Acode/GMAT/PKG/examples/data/yeast/CobaltChloride12'
id = 'ID'
trait = ['trait1', 'trait2']
agmat_file = '/nfs1/ningc/WORK/Acode/GMAT/PKG/examples/data/yeast/test'
varcom(data_file, id, trait, agmat_file, fix=None, init=None, maxiter=100, 
cc_par=1e-08, cc_gra=0.001, em_weight_step=0.01,
       prefix_outfile='/nfs1/ningc/WORK/Acode/GMAT/PKG/examples/data/yeast/mvlmm_varcom')






import sys
import pandas as pd
import numpy as np
from scipy import linalg
import logging
from scipy.sparse import csr_matrix

from ._common import _design_matrix


def _pre_data(data_file, id, trait, agmat_file, fix=None):
    logging.info('Data file: {}'.format(data_file))
    data_df = pd.read_csv(data_file, sep='\s+', header=0)
    col_names = data_df.columns
    logging.info('The column names of data file: {}'.format(col_names))
    logging.info('Note: Variates beginning with a capital letter is converted into factors.')
    class_vec = []
    for val in col_names:
        if not val[0].isalpha():
            logging.info("The first character of columns names must be alphabet!")
            sys.exit()
        if val[0] == val.capitalize()[0]:
            class_vec.append(val)
            data_df[val] = data_df[val].astype('str')
        else:
            try:
                data_df[val] = data_df[val].astype('float')
            except Exception as e:
                logging.info(e)
                logging.info("{} may contain string, please check!".format(val))
                sys.exit()
    if fix is None:
        fix = "1"
    fix_vec = []
    for val in list(col_names):
        if val in fix:
            fix_vec.append(val)
    id_in_data = []
    for val in trait:
        trait_df = data_df[[id] + fix_vec + [val]]
        ind = list(trait_df.isnull().any(axis=1) == False)
        trait_df = trait_df.iloc[ind, :]
        id_in_data.extend(list(trait_df[id]))
    data_df = data_df.sort_values(by=id)
    id_in_gmat = list(pd.read_csv(agmat_file + '.id', sep='\s+', header=None, dtype={0: 'str'}).iloc[:, 0])
    logging.info('Individual column: ' + id)
    if id not in class_vec:
        logging.info('The initial letter of {} should be capital'.format(id))
        sys.exit()
    id_order = []   # id in order
    ind_keep = []  # keep id both in phenotype and genotype file
    for val in data_df[id]:
        if (val in id_in_gmat) and (val in id_in_data):
            ind_keep.append(True)
            if val not in id_order:
                id_order.append(val)
        else:
            ind_keep.append(False)
    data_df = data_df.iloc[ind_keep, :]
    id_dct = dict(zip(id_order, range(len(id_order))))
    id_code = []  # code the id in the data file from 0 to ...
    for val in data_df[id]:
        id_code.append(id_dct[val])
    logging.info("The genomic relationship matrix file is: {}.agrm.id_fmt".format(agmat_file))
    ag = np.zeros((len(id_order), len(id_order)))
    with open(agmat_file + '.agrm.id_fmt') as fin:
        for line in fin:
            arr = line.split()
            try:
                ag[id_dct[arr[0]], id_dct[arr[1]]] = ag[id_dct[arr[1]], id_dct[arr[0]]] = float(arr[2])
            except Exception as _:
                pass
    logging.info('Trait column: {}'.format(trait))
    for val in trait:
        if val in class_vec:
            logging.error('The initial letter of {} should be lowercase'.format(trait))
            sys.exit()
    logging.info("The expression for the fixed effect: {}".format(fix))
    if fix is None:
        fix = "1"
    fix_vec = []
    for val in list(col_names):
        if val in fix:
            fix_vec.append(val)
    xmat_lst = []
    zmat_lst = []
    y_lst = []
    id_lst = []
    id_raw_lst = []
    for val in trait:
        trait_df = data_df[[id] + fix_vec + [val]]
        ind = list(trait_df.isnull().any(axis=1) == False)
        trait_df = trait_df.iloc[ind, :]
        # print(trait_df)
        y_lst.append(np.array(trait_df[val]).reshape(-1, 1))
        xmat_lst.append(_design_matrix(fix, trait_df))
        id_code_keep = np.array(id_code)[ind]
        len_row = len(id_code_keep)
        len_col = len(id_order)
        zmat = csr_matrix(([1.0]*len_row, (list(range(len_row)), id_code_keep)), shape=(len_row, len_col))
        zmat_lst.append(zmat)
        id_lst.append(id_code_keep)
        id_raw_lst.extend(list(trait_df[id]))

    def ind_cross(id1, id2):
        row = []
        col = []
        for i in range(len(id1)):
            try:
                col_i = id2.index(id1[i])
                col.append(col_i)
                row.append(i)
            except Exception as _:
                pass
        return row, col
    error_lst = []
    for i in range(len(id_lst)):
        error_vec = []
        for j in range(len(id_lst)):
            row, col = ind_cross(list(id_lst[i]), list(id_lst[j]))
            error = csr_matrix(([1.0]*len(row), (row, col)), shape=(len(id_lst[i]), len(id_lst[j])))
            error_vec.append(error)
        error_lst.append(error_vec)
    return y_lst, xmat_lst, zmat_lst, ag, error_lst, id_raw_lst



from gmat.uvlmm.gwas.fixed import fixed
from gmat.uvlmm.gwas.fixed_ import fixed_
from gmat.uvlmm.gwas.fixed_repeat import fixed_repeat

from gmat.uvlmm.gwas.fixed_gbye import fixed_gbye
from gmat.uvlmm.gwas.fixed_min import fixed_min



import logging
import sys
import numpy as np
from scipy.sparse import csr_matrix
from patsy import dmatrix
import sympy


def _leg_mt(time, tmax, tmin, order):
    time = np.array(time, dtype=float).reshape(-1, 1)
    tvec = 2 * (time - tmin) / (tmax - tmin) - 1
    pmat = []
    for k in range(order + 1):
        c = int(k / 2)
        j = k
        p = 0
        for r in range(0, c + 1):
            p += np.sqrt((2 * j + 1.0) / 2.0) * pow(0.5, j) * (pow(-1, r) *
                np.math.factorial(2 * j - 2 * r) / (np.math.factorial(r) *
                np.math.factorial(j - r) * np.math.factorial(j - 2 * r))) * pow(tvec, j - 2 * r)
        p = np.array(p)
        pmat.append(p)
    return pmat


def _design_matrix(formula, data_df):
    """
    Build the design matrix
    :param formula:  An object that can be used to construct a design matrix. See patsy.dmatrix for detail.
    :param data_df: Pandas data frame
    :return: full rank design matrix
    """
    dmat = dmatrix(formula, data_df)
    dmat = np.asarray(dmat)
    return dmat


def _design_matrix_crank(formula, data_df):
    """
    Build the design matrix
    :param formula:  An object that can be used to construct a design matrix. See patsy.dmatrix for detail.
    :param data_df: Pandas data frame
    :return: design matrix, full rank design matrix, the indexes for the linear independent columns, label of design matrix
    """
    dmat = dmatrix(formula, data_df)
    col_arr = repr(dmat).split("\n")[1].split('\s+')
    dmat = np.asarray(dmat)
    # _, indexes = sympy.Matrix(dmat).rref()
    _, r = np.linalg.qr(dmat)
    indexes = np.absolute(np.diag(r)) >= 1e-10
    dmat_full_rank = dmat[:, indexes]
    return dmat, dmat_full_rank, indexes, col_arr

def _full_col_rank(mat):
    _, r = np.linalg.qr(mat)
    indexes = np.absolute(np.diag(r)) >= 1e-10
    mat_full_rank = mat[:, indexes]
    return mat_full_rank



"""
Calculate the additive genomic relationship matrix


"""

import logging
import os
from gmat.gmatrix import agmat
from gmat.gmatrix import agmat_dosage
from gmat.gmatrix import pca


logging.basicConfig(level=logging.INFO)

bed_file = '../data/mouse/plink'
out_file = '../data/mouse/test2'

# matrix form
#agmat0 = agmat(bed_file, out_file, inv=True, small_val=0.001, out_fmt='mat')

# row-column-value form, which can be used by asreml

# id-id-value form
agmat2 = agmat(bed_file, out_file, inv=False, small_val=0.001, out_fmt='id_id_val')

"""
out_file = '../data/mouse/test2'
geno_file = '/nfs1/ningc/WORK/Acode/GMAT/PKG/examples/data/mouse/geno.txt'
agmat_dosage(geno_file, out_file, inv=True, out_fmt='id_id_val', npart=10, small_val=0.001)
"""

# PCA and heatmap
pca("../data/mouse/test")




import logging
from gmat.gmatrix import agmat
from gmat.uvlmm.gwas import fixed, fixed_gbye
logging.basicConfig(level=logging.INFO)
pheno_file = 'pheno'
bed_file = 'plink'


agmat_file = 'test'
agmat2 = agmat(bed_file, out_file=agmat_file, inv=False, small_val=0.001, out_fmt='id_id_val')

fixed(pheno_file, bed_file, agmat_file, class_lst=["mean", "sex", "treat"], fix="mean + sex + age + treat", out_file='res.txt', 
     condition_snp=None, npart=100, maxiter0=100, maxiter1=10, speed=True, pcut=5e-05)


# 基因环境互作、不加永久环境效应，batch是连续变量
bye = 'age'
fixed_gbye(pheno_file, bed_file, agmat_file, bye, class_lst=["mean", "sex", "treat"], fix="mean + sex + age + treat",
           out_file='res3.txt', condition_snp=None, npart=100,
          maxiter0=100, maxiter1=10, speed=True, pcut=5.0e-5)

# 基因环境互作、不加永久环境效应，batch是分类变量
bye = 'age'
fixed_gbye(pheno_file, bed_file, agmat_file, bye, class_lst=["mean", "sex", "treat"], fix="mean + sex + age + treat",
           out_file='res4.txt', condition_snp=None, npart=100,
          maxiter0=100, maxiter1=10, speed=True, pcut=5.0e-5)



import numpy as np
import pandas as pd
from scipy import linalg
import logging
import sys
import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
import seaborn as sns


def pca(agmat_file, pca_num=10, keep_file=None):
    """
    Principal component analysis
    :param agmat_file: the prefix of file for the additive genomic relationship matrix from the agmat function (id_id_val)
    :param pca_num: the number of pca to output. Default is 10.
    :param keep_file: the file for a subset of individuals. Default is None, all the individuals will be used.
    :return: the eigenvectors (*.eigenvec) and  all the eigenvalues (*.eigenval)
    """
    logging.info("###Read the relationship matrix with given IDs###")
    id_agmat = np.array(pd.read_csv(agmat_file + '.id', header=None, dtype=np.str).iloc[:, 0])
    id_keep = np.array(id_agmat)
    if keep_file is not None:
        id_keep = np.array(pd.read_csv(keep_file, header=None, dtype=np.str).iloc[:, 0])
        id_del = set(id_keep) - set(id_agmat)
        if len(id_del) != 0:
            logging.error('These individuals are not in the agmat_file: {}'.format(id_del))
            sys.exit()
    id_dct = dict(zip(id_keep, range(len(id_keep))))
    ag = np.zeros((len(id_keep), len(id_keep)))
    with open(agmat_file + '.agrm.id_fmt') as fin:
        for line in fin:
            arr = line.split()
            try:
                ag[id_dct[arr[0]], id_dct[arr[1]]] = ag[id_dct[arr[1]], id_dct[arr[0]]] = float(arr[2])
            except Exception as e:
                del e
    logging.info('###Perform the eigendecomposition###')
    eigenvals, eigenvecs = linalg.eigh(ag)
    eigenvals = np.flip(eigenvals)
    all_sum = np.sum(eigenvals)
    eigenvals_ratio = eigenvals / all_sum
    eigenvals_accum = []
    accum_sum = 0
    for val in eigenvals:
        accum_sum += val
        eigenvals_accum.append(accum_sum / all_sum)
    eigenvals_df = pd.DataFrame({"val": eigenvals, "ratio": eigenvals_ratio, "accum": eigenvals_accum})
    eigenvals_df.to_csv(agmat_file + '.eigenvals', sep=' ', header=False, index=False)
    eigenvecs_df = pd.concat([pd.DataFrame({'id': id_keep}), pd.DataFrame(eigenvecs[:, range(-1, -pca_num-1, -1)])], axis=1)
    eigenvecs_df.to_csv(agmat_file + '.eigenvecs', sep=' ', header=False, index=False)
    logging.info("###2D plot with the first two PCs###")
    plt.figure(figsize=(3.5, 3.5), dpi=600)
    plt.rc('font', family='Times New Roman')
    plt.xlabel('PC1', fontsize=10)
    plt.ylabel('PC2', fontsize=10)
    plt.xticks(fontsize=8)
    plt.yticks(fontsize=8)
    plt.scatter(eigenvecs[:, -1], eigenvecs[:, -2], s=1.2, c='black')
    plt.tight_layout()
    plt.savefig(agmat_file + '.pca2D.png')
    logging.info("###3D plot with the first two PCs###")
    plt.figure(figsize=(3.5, 3.5), dpi=600)
    plt.rc('font', family='Times New Roman')
    ax = plt.axes(projection='3d')
    ax.set_xlabel('PC1', fontsize=10)
    ax.set_ylabel('PC2', fontsize=10)
    # ax.set_zlabel('PC3', fontsize=10)
    ax.xaxis.set_tick_params(labelsize=8)
    ax.yaxis.set_tick_params(labelsize=8)
    ax.zaxis.set_tick_params(labelsize=8)
    ax.scatter(eigenvecs[:, -1], eigenvecs[:, -2], eigenvecs[:, -3], s=1.2, c='black')
    ax.set_in_layout(True)
    plt.savefig(agmat_file + '.pca3D.png')
    logging.info('###Heatmap for the relationship###')
    plt.figure(figsize=(4, 3.5), dpi=600)
    plt.rc('font', family='Times New Roman')
    sns.set()
    sns.heatmap(ag, cmap='YlGnBu', xticklabels=False, yticklabels=False)
    cax = plt.gcf().axes[-1]
    cax.tick_params(labelsize=10, direction='in', top='off', bottom='off', left='off', right='off')
    plt.savefig(agmat_file + '.heatmap.png')
    return 0



import os
import logging
from scipy import linalg
from scipy.sparse import csr_matrix
import numpy as np
import pandas as pd
import gc

from ._common import _design_matrix, _leg_mt, _full_col_rank


def _pre_data(data_file, id, tpoint, trait, gmat_file, tfix=None, fix=None, forder=3, rorder=3, na_method='omit',
                   init=None):
    logging.info('***Read the data file***')
    logging.info('Data file: {}'.format(data_file))
    data_df = pd.read_csv(data_file, sep='\s+', header=0)
    logging.info('NA method: ' + na_method)
    if na_method == 'omit':
        data_df = data_df.dropna()
    elif na_method == 'include':
        data_df = data_df.fillna(method='ffill')
        data_df = data_df.fillna(method='bfill')
    else:
        logging.info('na_method does not exist: {}'.format(na_method))
        exit()
    col_names = data_df.columns
    logging.info('The column names of data file: {}'.format(list(col_names)))
    logging.info('Note: Variates beginning with a capital letter is converted into factors.')
    class_vec = []
    for val in col_names:
        if not val[0].isalpha():
            logging.info("The first character of columns names must be alphabet!")
            exit()
        if val[0] == val.capitalize()[0]:
            class_vec.append(val)
            data_df[val] = data_df[val].astype('str')
        else:
            try:
                data_df[val] = data_df[val].astype('float')
            except Exception as e:
                logging.info(e)
                logging.info(val + " column may contain string, please check!")
                exit()
    logging.info('Individual column: {}'.format(id))
    if id not in class_vec:
        logging.info('The initial letter of {} should be capital'.format(id))
        exit()
    id_in_data = list(data_df[id])
    if len(set(id_in_data)) != len(id_in_data):
        logging.info('Duplicated ids exit in the data file, please check!')
        exit()
    id_in_gmat = list(pd.read_csv(gmat_file + '.id', sep='\s+', header=None, dtype={0: 'str'}).iloc[:, 0])
    ind_keep = []
    for val in data_df[id]:
        if val in id_in_gmat:
            ind_keep.append(True)
        else:
            ind_keep.append(False)
    data_df = data_df.iloc[ind_keep, :]
    logging.info('Trait column name: {}'.format(trait))
    if len(set(trait) & set(class_vec)) != 0:
        logging.info('Phenotype should not be defined as class variable, please check!')
        exit()
    logging.info("The genomic relationship matrix file is: {}.agrm.id_fmt".format(gmat_file))
    id_dct = dict(zip(id_in_data, range(len(id_in_data))))
    ag = np.zeros((len(id_in_data), len(id_in_data)))
    with open(gmat_file + '.agrm.id_fmt') as fin:
        for line in fin:
            arr = line.split()
            try:
                ag[id_dct[arr[0]], id_dct[arr[1]]] = ag[id_dct[arr[1]], id_dct[arr[0]]] = float(arr[2])
            except Exception as e:
                del e
    logging.info('***Eigen decomposition of kinship matrix***')
    ag_eigen_val, ag_eigen_vec = linalg.eigh(ag)
    logging.info('***Build the design matrix for fixed effect***')
    tmin = np.min(tpoint)
    tmax = np.max(tpoint)
    leg_fix = _leg_mt(np.array(tpoint), tmax, tmin, forder)
    xmat_bt = np.concatenate(leg_fix, axis=1)
    xmat_t = [xmat_bt] * data_df.shape[0]
    if tfix is not None:
        xmat_t = []
        dmat_t = _design_matrix(tfix, data_df)
        for i in range(dmat_t.shape[0]):
            mat = []
            for j in range(dmat_t.shape[1]):
                mat.append(xmat_bt * dmat_t[i, j])
            mat = np.concatenate(mat, axis=1)
            xmat_t.append(mat)
    xmat = np.concatenate(xmat_t, axis=0)
    if fix is not None:
        xmat_nt = []
        dmat_nt = _design_matrix(fix, data_df)
        for i in range(dmat_nt.shape[0]):
            mat = []
            for j in range(1, dmat_nt.shape[1]):
                mat.append(np.array([dmat_nt[i, j]] * len(tpoint)).reshape(-1, 1))
            mat = np.concatenate(mat, axis=1)
            xmat_nt.append(mat)
        xmat_nt = np.concatenate(xmat_nt, axis=0)
        xmat = np.concatenate([xmat, xmat_nt], axis=1)
    if np.linalg.matrix_rank(xmat) < xmat.shape[1]:
        xmat = _full_col_rank(xmat)
    xmat = np.dot(linalg.kron(ag_eigen_vec.T, np.eye(len(tpoint))), xmat)
    xmat = xmat.reshape(data_df.shape[0], len(tpoint), -1)
    logging.info('***Initial values for variances and rotate the phenotypic values')
    leg_tp = _leg_mt(np.array(tpoint), tmax, tmin, rorder)
    leg_tp = np.concatenate(leg_tp, axis=1)
    y = np.array(data_df[trait])
    cov_dim = leg_tp.shape[1]
    y_var = np.var(y) / (cov_dim * 2 + 1)
    var_com = []
    cov_var = np.diag([y_var] * cov_dim)
    var_com.extend(cov_var[np.tril_indices_from(cov_var)])
    var_com.extend(cov_var[np.tril_indices_from(cov_var)])
    var_com.append(y_var)
    if init is None:
        var_com = np.array(var_com)
    else:
        if len(var_com) != len(init):
            logging.info('The length of initial variances should be {}'.format(len(var_com)))
            exit()
        else:
            var_com = np.array(init)
    y = np.dot(ag_eigen_vec.T, y)
    y = y.reshape(data_df.shape[0], len(tpoint), 1)
    return y, xmat, leg_tp, var_com, ag_eigen_val



import numpy as np
from scipy import linalg
from functools import reduce
import logging
from scipy.sparse import hstack


def ai_vmat(y, xmat, zmat_lst, gmat_lst, init=None, maxiter=100, cc_par=1.0e-8):
    """
    :param y:
    :param xmat:
    :param zmat_lst:
    :param gmat_lst:
    :param init:
    :param maxiter:
    :param cc_par:
    :return:
    """
    logging.info("#####Prepare#####")
    num_var = len(gmat_lst) + 1
    var_com = np.ones(num_var)
    if init is not None:
        var_com = np.array(init)
    y = np.array(y).reshape(-1, 1)
    n = y.shape[0]
    xmat = np.array(xmat).reshape(n, -1)
    zgzmat_lst = []
    for val in range(len(gmat_lst)):
        zgzmat_lst.append(zmat_lst[val].dot((zmat_lst[val].dot(gmat_lst[val])).T))
    logging.info('Initial variances: ' + ' '.join(list(np.array(var_com, dtype=str))))
    logging.info("#####Start the iteration#####\n\n")
    iter = 0
    cc_par_val = 1000.0
    while iter < maxiter:
        iter += 1
        logging.info('Round: {:d}'.format(iter))
        # V, the inverse of V and P
        vmat = np.diag([var_com[-1]] * n)
        for val in range(len(zgzmat_lst)):
            vmat += zgzmat_lst[val] * var_com[val]
        vmat = np.linalg.inv(vmat)
        vxmat = np.dot(vmat, xmat)
        xvxmat = np.dot(xmat.T, vxmat)
        xvxmat = linalg.inv(xvxmat)
        pmat = vmat - reduce(np.dot, [vxmat, xvxmat, vxmat.T])
        pymat = np.dot(pmat, y)
        # first partial derivatives and working variates
        fd_mat = np.zeros(num_var)
        wv_mat = []
        fd_mat[-1] = np.trace(pmat) - np.sum(np.dot(pymat.T, pymat))
        fd_mat[-1] = -0.5 * fd_mat[-1]
        for i in range(num_var - 1):
            fd_mat[i] = np.sum(pmat*zgzmat_lst[i]) - np.sum(reduce(np.dot, [pymat.T, zgzmat_lst[i], pymat]))
            fd_mat[i] = -0.5 * fd_mat[i]
            wv_mat.append(np.dot(zgzmat_lst[i], pymat))
        wv_mat.append(pymat)
        wv_mat = np.concatenate(wv_mat, axis=1)
        ai_mat = 0.5 * reduce(np.dot, [wv_mat.T, pmat, wv_mat])
        delta = np.dot(linalg.inv(ai_mat), fd_mat)
        var_com_new = var_com + delta
        cc_par_val = np.sum(delta * delta) / np.sum(np.array(var_com_new) * np.array(var_com_new))
        cc_par_val = np.sqrt(cc_par_val)
        var_com = np.array(var_com_new)
        logging.info('Norm of update vector: {:e}'.format(cc_par_val))
        var_com_str = ' '.join(np.array(var_com, dtype=str))
        logging.info('Updated variances: {}'.format(var_com_str))
        if cc_par_val < cc_par:
            break
    if cc_par_val < cc_par:
        logging.info('Variances converged.')
    else:
        logging.info('Variances not converged.')
    return var_com


def ai_mme(y, xmat, zmat_lst, gmat_inv_lst, init=None, maxiter=100, cc_par=1.0e-8):
    """
    Estimate variance parameters with mme based em
    :param y:
    :param xmat:
    :param gmat_inv_lst:
    :param init:
    :param maxiter:
    :param cc_par:
    :param cc_gra:
    :return:
    """
    num_var = len(gmat_inv_lst) + 1
    var_com = np.ones(num_var)
    if init is not None:
        var_com = np.array(init)
    y = np.array(y).reshape(-1, 1)
    n = y.shape[0]
    xmat = np.array(xmat).reshape(n, -1)
    xmat_df = xmat.shape[1]
    df_index = [xmat_df]  # the index for the degree of freedom for different effects.
    for i in range(num_var - 1):
        df_index.append(df_index[-1] + gmat_inv_lst[i].shape[0])
    logging.info("Prepare the coefficient matrix without variance parameters")
    zmat_concat = hstack(zmat_lst)
    xzmat = hstack([xmat, zmat_concat])
    coef_null = xzmat.T.dot(xzmat)  # the coefficient matrix
    rhs_null = xzmat.T.dot(y)  # the right hand matrix
    logging.info('Initial variances: ' + ' '.join(list(np.array(var_com, dtype=str))))
    logging.info("Start iteration")
    iter = 0
    cc_par_val = 10000.0
    while iter < maxiter:
        iter += 1
        logging.info('Round: {:d}'.format(iter))
        # coef matrix
        coef = coef_null.toarray() / var_com[-1]
        for k in range(num_var - 1):
            indexa = df_index[k]
            indexb = df_index[k + 1]
            coef[indexa:indexb, indexa:indexb] = coef[indexa:indexb, indexa:indexb] + gmat_inv_lst[k] / var_com[k]
        # right hand
        rhs_mat = rhs_null / var_com[-1]
        coef_inv = linalg.inv(coef)
        eff = np.dot(coef_inv, rhs_mat)
        e_hat = y - xzmat.dot(eff)
        # first partial derivatives and working variates
        fd_mat = np.zeros(num_var)
        wv_mat = []
        for i in range(num_var - 1):
            indexa = df_index[i]
            indexb = df_index[i + 1]
            fd_mat[i] = np.sum(coef_inv[indexa:indexb, indexa:indexb] * gmat_inv_lst[i]) + \
                        np.sum(reduce(np.dot, [eff[indexa:indexb, :].T, gmat_inv_lst[i], eff[indexa:indexb, :]]))
            fd_mat[i] = -0.5 * gmat_inv_lst[i].shape[0]/var_com[i] + 0.5 * fd_mat[i] / (var_com[i] * var_com[i])
            wv_mati = zmat_lst[i].dot(eff[indexa:indexb, :]) / var_com[i]
            wv_mat.append(wv_mati)
        fd_mat[-1] = np.sum((xzmat.T.dot(xzmat)).multiply(coef_inv)) + np.sum(np.dot(e_hat.T, e_hat))
        fd_mat[-1] = -0.5*n/var_com[-1] + 0.5 * fd_mat[-1] / (var_com[-1] * var_com[-1])
        wv_mati = e_hat / var_com[-1]
        wv_mat.append(wv_mati)
        wv_mat = np.concatenate(wv_mat, axis=1)
        wvxz_mat = xzmat.T.dot(wv_mat)/var_com[-1]
        ai_mat = np.dot(wv_mat.T, wv_mat) / var_com[-1] - reduce(np.dot, [wvxz_mat.T, coef_inv, wvxz_mat])
        ai_mat = 0.5 * ai_mat
        delta = np.dot(linalg.inv(ai_mat), fd_mat)
        var_com_new = var_com + delta
        cc_par_val = np.sum(delta * delta) / np.sum(np.array(var_com_new) * np.array(var_com_new))
        cc_par_val = np.sqrt(cc_par_val)
        var_com = np.array(var_com_new)
        logging.info('Norm of update vector: {:e}'.format(cc_par_val))
        var_com_str = ' '.join(np.array(var_com, dtype=str))
        logging.info('Updated variances: {}'.format(var_com_str))
        if cc_par_val < cc_par:
            break
    if cc_par_val < cc_par:
        logging.info('Variances converged.')
    else:
        logging.info('Variances not converged.')
    return var_com



import logging
import numpy as np
import pandas as pd
from scipy import linalg
from pysnptools.snpreader import Bed
from tqdm import tqdm
from gmat.process_plink.process_plink import impute_geno
from scipy.stats import chi2
from scipy.sparse import csc_matrix
from patsy import dmatrix
from collections import OrderedDict
from functools import reduce
from scipy.stats import chi2


def weight_emai_eigen(pheno_file, agmat_file, class_lst=None, fix=None, maxiter0=100):
    """
    run GWAS based on linear mixed model and put one snp in each analysis.
    :param pheno_file: phenotypic file with head rows. The fist column is individual id. The last column is phenotypic values.
    Class variates and covariates are in the middle columns. Missing values in the phenotypic file can be
    expressed as 'NA', 'NaN', 'nan', 'na'.
    :param agmat_file: the prefix for the additive genomic relationship matrix file.
    :param class_lst: a list for class variates
    :param fix: formula for the fix effect. Eg. fix='sex + herd'
    :param maxiter0: the maximum number of iteration for variance estimation in the null model
    :return:
    """
    logging.info("###Prepare the phenotypic values, covariates and genomic matrix")
    y, xmat, zmat, ag, id_keep, xmat_ind = _pre_pheno(pheno_file, agmat_file, class_lst, fix)
    logging.info("###Estimate the variances in the null model.")
    init = np.array([np.var(y) / 2] * 2)
    gmat_eigenvals, gmat_eigenvecs = linalg.eigh(zmat.dot((zmat.dot(ag)).T))
    gmat_eigenvals = np.array(gmat_eigenvals).reshape(-1, 1)
    y_trans = np.dot(gmat_eigenvecs.T, y)  # transform the phenotypic values
    xmat_trans = np.dot(gmat_eigenvecs.T, xmat)  # transform the fixed effects
    var, convergence = _weight_emai_eigen(y_trans, xmat_trans, gmat_eigenvals, init=init, maxiter=maxiter0, cc=1.0e-8)
    logging.info('Whether converge?: {}'.format(convergence))
    logging.info('The estimated variances: {}'.format(var))
    logging.info("###Test the significance of fixed effects###")
    vmat = 1.0 / (gmat_eigenvals * var[0] + var[1])
    vx = np.multiply(vmat, xmat_trans)
    xvx = np.dot(xmat_trans.T, vx)
    xvx = linalg.inv(xvx)
    xvy = np.dot(vx.T, y_trans)
    eff = np.dot(xvx, xvy)
    for key, value in xmat_ind.items():
        logging.info("***{}: ".format(key))
        logging.info("Effect: {}".format(eff[value, -1]))
        logging.info("Covariance: {}".format(xvx[value, value]))
        p_val = chi2.sf(np.sum(reduce(np.dot, [eff[value, :].T, linalg.inv(xvx[value, :][:, value]), eff[value, :]])), len(value))
        logging.info("P value: {}".format(p_val))


def _pre_pheno(pheno_file, agmat_file, class_lst, fix):
    id_in_agmat = []
    with open(agmat_file + '.id') as fin:
        for line in fin:
            arr = line.split()
            id_in_agmat.append(arr[0])
    data_df = pd.read_csv(pheno_file, sep='\s+', header=0)
    data_df = data_df.dropna()
    col_names = data_df.columns
    class_lst2 = [col_names[0]]
    if class_lst is not None:
        class_lst2 = class_lst + class_lst2
    for val in col_names:
        if val in class_lst2:
            data_df[val] = data_df[val].astype('str')
        else:
            data_df[val] = data_df[val].astype('float')
    id_keep = []  # keep the id both in the agmat file and pheno file, without missing phenotypes and covariates.
    for i in range(data_df.shape[0]):
        if data_df.iloc[i, 0] in id_in_agmat:
            id_keep.append(True)
        else:
            id_keep.append(False)
    data_df = data_df.iloc[id_keep, :]
    id_keep = list(data_df.iloc[:, 0])
    y = np.array(data_df.iloc[:, -1]).reshape(-1, 1)
    xmat = np.ones((y.shape[0], 1))
    xmat_ind = OrderedDict()
    xmat_ind['Intercept'] = [0]
    if fix is not None:
        xmat = dmatrix(fix, data_df)
        xcol_names = xmat.design_info.column_names
        xmat = np.asarray(xmat)
        _, r = np.linalg.qr(xmat)
        indexes = np.absolute(np.diag(r)) >= 1e-10
        xcol_names = np.array(xcol_names)[indexes]
        for i in range(1, len(xcol_names)):
            key = xcol_names[i].split('[')[0]
            try:
                xmat_ind[key].append(i)
            except Exception as _:
                xmat_ind[key] = [i]
        xmat = xmat[:, indexes]
    val = 0
    id_dct = {}
    col = []
    for v in id_keep:
        if v not in id_dct:
            id_dct[v] = val
            val += 1
        col.append(id_dct[v])
    row = range(len(id_keep))
    zmat = csc_matrix(([1.0]*len(row), (row, col)))
    ag = np.zeros((len(id_dct), len(id_dct)))
    with open(agmat_file + '.agrm.id_fmt') as fin:
        for line in fin:
            arr = line.split()
            try:
                ag[id_dct[arr[0]], id_dct[arr[1]]] = ag[id_dct[arr[1]], id_dct[arr[0]]] = float(arr[2])
            except Exception as e:
                del e
    return y, xmat, zmat, ag, id_keep, xmat_ind


def _weight_emai_eigen(y_trans, xmat_trans, gmat_eigenvals, init=None, maxiter=100, cc=1.0e-8):
    fd_mat = np.zeros(2)
    ai_mat = np.zeros((2, 2))
    em_mat = np.zeros((2, 2))
    num_id = gmat_eigenvals.shape[0]
    cc_val = 1000.0
    var = np.array(init)
    var_update = var * 1000
    delta = var_update - var
    convergence = False
    for i in range(maxiter):
        logging.info("Iteration: {}".format(i+1))
        vmat = 1.0 / (gmat_eigenvals * var[0] + var[1])
        vx = np.multiply(vmat, xmat_trans)
        xvx = np.dot(xmat_trans.T, vx)
        xvx = linalg.inv(xvx)
        # py
        xvy = np.dot(vx.T, y_trans)
        y_xb = y_trans - np.dot(xmat_trans, np.dot(xvx, xvy))
        py = np.multiply(vmat, y_xb)
        # add_py p_add_py
        add_py = np.multiply(gmat_eigenvals, py)
        xvy = np.dot(vx.T, add_py)
        y_xb = add_py - np.dot(xmat_trans, np.dot(xvx, xvy))
        p_add_py = np.multiply(vmat, y_xb)
        # res_py p_res_py
        res_py = py.copy()
        xvy = np.dot(vx.T, res_py)
        y_xb = res_py - np.dot(xmat_trans, np.dot(xvx, xvy))
        p_res_py = np.multiply(vmat, y_xb)
        # fd
        tr_vd = np.sum(np.multiply(vmat, gmat_eigenvals))
        xvdvx = np.dot(xmat_trans.T, vmat * gmat_eigenvals * vx)
        tr_2d = np.sum(np.multiply(xvdvx, xvx))
        ypvpy = np.sum(np.dot(py.T, add_py))
        fd_mat[0] = 0.5 * (-tr_vd + tr_2d + ypvpy)
        tr_vd = np.sum(vmat)
        xvdvx = np.dot(xmat_trans.T, vmat * vx)
        tr_2d = np.sum(np.multiply(xvdvx, xvx))
        ypvpy = np.sum(np.dot(py.T, res_py))
        fd_mat[1] = 0.5 * (-tr_vd + tr_2d + ypvpy)
        # AI
        ai_mat[0, 0] = np.sum(np.dot(add_py.T, p_add_py))
        ai_mat[0, 1] = ai_mat[1, 0] = np.sum(np.dot(add_py.T, p_res_py))
        ai_mat[1, 1] = np.sum(np.dot(res_py.T, p_res_py))
        ai_mat = 0.5 * ai_mat
        # EM
        em_mat[0, 0] = num_id / (2 * var[0] * var[0])
        em_mat[1, 1] = num_id / (2 * var[1] * var[1])
        for j in range(0, 51):
            gamma = j * 0.02
            wemai_mat = (1 - gamma) * ai_mat + gamma * em_mat
            delta = np.dot(linalg.inv(wemai_mat), fd_mat)
            var_update = var + delta
            if min(var_update) > 0:
                logging.info('EM weight value: {}'.format(gamma))
                break
        # Convergence criteria
        logging.info('Updated variances: {}'.format(var_update))
        cc_val = np.sum(pow(delta, 2)) / np.sum(pow(var_update, 2))
        cc_val = np.sqrt(cc_val)
        var = var_update.copy()
        logging.info("CC: {}".format(cc_val))
        if cc_val < cc:
            break
    if cc_val < cc:
        convergence = True
    return var, convergence



import logging
from functools import reduce
import numpy as np
import pandas as pd
from scipy import linalg
from pysnptools.snpreader import Bed
from tqdm import tqdm
from gmat.process_plink.process_plink import impute_geno
from scipy.stats import chi2
from scipy.sparse import csc_matrix
from patsy import dmatrix


def fixed_repeat(pheno_file, bed_file, agmat_file, bye=None, class_lst=None, fix=None, out_file='res.txt', condition_snp=None, npart=100,
          maxiter=100):
    """
    run GWAS based on linear mixed model and put one snp in each analysis.
    :param pheno_file: phenotypic file with head rows. The fist column is individual id. The last column is phenotypic values.
    Class variates and covariates are in the middle columns. Missing values in the phenotypic file can be
    expressed as 'NA', 'NaN', 'nan', 'na'.
    :param bed_file: the prefix for the plink binary file.
    :param agmat_file: the prefix for the additive genomic relationship matrix file.
    :param bye: the environment effect that SNP interacts with. Default is None.
    :param class_lst: a list for class variates
    :param fix: formula for the fix effect. Eg. fix='sex + herd'
    :param out_file: the output file. Default is 'res.txt'.
    :param condition_snp: A string or a list for the conditional SNP. Default is None.
    :param npart: divide the SNPs into n parts and read them into memory successively. Default is 100.
    :param maxiter: the maximum number of iteration for variance estimation in the null model.
    :return:
    """
    logging.info("###Prepare the phenotypic values, covariates and genomic matrix")
    y, xmat, zmat, ag, id_keep, bye_mat = _pre_pheno(pheno_file, agmat_file, bye, class_lst, fix)
    logging.info("###Estimate the variances in the null model.")
    init = np.array([np.var(y) / 3] * 3)
    var, convergence, vmat = _weight_emai_vmat(y, xmat, zmat, ag, init=init, maxiter=maxiter, cc_par=1.0e-8)
    vmat = linalg.inv(vmat)
    logging.info('Whether converge?: {}'.format(convergence))
    logging.info('The estimated variances: {}'.format(var))
    snp_on_disk = Bed(bed_file, count_A1=False)
    num_snp = snp_on_disk.sid_count
    fam_df = pd.read_csv(bed_file + '.fam', header=None, sep='\s+')
    id_in_fam = list(np.array(fam_df.iloc[:, 1], dtype=np.str))
    id_keep_ind = []
    for val in id_keep:
        id_keep_ind.append(id_in_fam.index(val))
    snp_index = []
    for i in range(npart):
        snp_index.append(int(num_snp / npart) * i)
    snp_index.append(num_snp)
    logging.info("Conditional SNP: {}".format(condition_snp))
    if condition_snp is not None:
        if not isinstance(condition_snp, list):
            condition_snp = [condition_snp]
        condition_snp_index = snp_on_disk.sid_to_index(condition_snp)
        condition_snp_val = snp_on_disk[id_keep_ind, condition_snp_index].read().val
        if np.any(np.isnan(condition_snp_val)):
            logging.info('Missing genotypes of conditional SNP are imputed with random genotypes.')
            condition_snp_val = impute_geno(condition_snp_val)
        xmat_con = np.array(condition_snp_val)
        if bye_mat is not None:
            xmat_con = np.concatenate([condition_snp_val, condition_snp_val*bye_mat], axis=1)
        xmat = np.concatenate([xmat, xmat_con], axis=1)
    logging.info('###Start the association')
    with open(out_file, 'w') as fout:
        if bye_mat is not None:
            fout.write('\t'.join(['chro', 'snp_ID', 'pos', 'allele1', 'allele2', 'eff', 'se', 'p_snp', 'p_gbye', 'p_all']))
            fout.write('\n')
        else:
            fout.write('\t'.join(['chro', 'snp_ID', 'pos', 'allele1', 'allele2', 'eff', 'se', 'p_snp']))
            fout.write('\n')
    for i in range(npart):
        logging.info('The {}/{} part'.format(i + 1, npart))
        snp_info = pd.read_csv(bed_file + '.bim', sep='\s+', header=None, skiprows=snp_index[i],
                               nrows=snp_index[i + 1] - snp_index[i])
        res_df = snp_info.iloc[:, [0, 1, 3, 4, 5]]
        res_df.columns = ['chro', 'snp_ID', 'pos', 'allele1', 'allele2']
        snp_mat = snp_on_disk[id_keep_ind, snp_index[i]:snp_index[i + 1]].read().val
        if np.any(np.isnan(snp_mat)):
            snp_mat = impute_geno(snp_mat)
        eff_vec = []
        se_vec = []
        p_snp_vec = []
        p_gbye_vec = []
        p_all_vec = []
        if bye_mat is not None:
            bye_df = bye_mat.shape[1]
            for k in tqdm(range(snp_mat.shape[1])):
                snpk = np.concatenate([snp_mat[:, k:(k + 1)], snp_mat[:, k:(k + 1)] * bye_mat], axis=1)
                xmat_k = np.concatenate([xmat, snpk], axis=1)
                # main
                vx = np.dot(vmat, xmat_k[:, :-bye_df])
                xvx = np.dot(xmat_k[:, :-bye_df].T, vx)
                try:
                    xvx = linalg.inv(xvx)
                except Exception as _:
                    xvx = np.diag(np.diag(xvx))*0.001 + xvx
                    xvx = linalg.inv(xvx)
                xvy = np.dot(vx.T, y)
                b_k = np.dot(xvx, xvy)
                eff = b_k[-1, -1]
                se = np.sqrt(xvx[-1, -1])
                t_val = eff / se
                p_val = chi2.sf(t_val * t_val, 1)
                # GbyE
                vx = np.dot(vmat, xmat_k)
                xvx = np.dot(xmat_k.T, vx)
                try:
                    xvx = linalg.inv(xvx)
                except Exception as _:
                    xvx = np.diag(np.diag(xvx))*0.001 + xvx
                    xvx = linalg.inv(xvx)
                xvy = np.dot(vx.T, y)
                b_k = np.dot(xvx, xvy)
                eff_bye = b_k[-bye_df:, :]
                eff_bye_cov = xvx[-bye_df:, -bye_df:]
                chi_val = np.sum(np.dot(eff_bye.T, np.dot(linalg.inv(eff_bye_cov), eff_bye)))
                p_bye = chi2.sf(chi_val, bye_df)
                chi_all = np.sum(np.dot(b_k[-(bye_df + 1):, :].T,
                                        np.dot(linalg.inv(xvx[-(bye_df + 1):, -(bye_df + 1):]),
                                               b_k[-(bye_df + 1):, :])))
                p_all = chi2.sf(chi_all, bye_df + 1)
                eff_vec.append(eff)
                se_vec.append(se)
                p_snp_vec.append(p_val)
                p_gbye_vec.append(p_bye)
                p_all_vec.append(p_all)
            res_df.loc[:, 'eff_val'] = eff_vec
            res_df.loc[:, 'se_val'] = se_vec
            res_df.loc[:, 'p_snp_val'] = p_snp_vec
            res_df.loc[:, 'p_gbye_val'] = p_gbye_vec
            res_df.loc[:, 'p_all_val'] = p_all_vec
        else:
            for k in tqdm(range(snp_mat.shape[1])):
                xmat_k = np.concatenate([xmat, snp_mat[:, k:(k + 1)]], axis=1)
                vx = np.dot(vmat, xmat_k)
                xvx = np.dot(xmat_k.T, vx)
                try:
                    xvx = linalg.inv(xvx)
                except Exception as _:
                    xvx = np.diag(np.diag(xvx)) * 0.001 + xvx
                    xvx = linalg.inv(xvx)
                xvy = np.dot(vx.T, y)
                b_k = np.dot(xvx, xvy)
                eff = b_k[-1, -1]
                se = np.sqrt(xvx[-1, -1])
                t_val = eff / se
                p_val = chi2.sf(t_val * t_val, 1)
                eff_vec.append(eff)
                se_vec.append(se)
                p_snp_vec.append(p_val)
            res_df.loc[:, 'eff_val'] = eff_vec
            res_df.loc[:, 'se_val'] = se_vec
            res_df.loc[:, 'p_snp_val'] = p_snp_vec
        res_df.to_csv(out_file, sep='\t', index=False, header=False, mode='a', na_rep='NA')
    return 0


def _pre_pheno(pheno_file, agmat_file, bye, class_lst, fix):
    id_in_agmat = []
    with open(agmat_file + '.id') as fin:
        for line in fin:
            arr = line.split()
            id_in_agmat.append(arr[0])
    data_df = pd.read_csv(pheno_file, sep='\s+', header=0)
    data_df = data_df.dropna()
    col_names = data_df.columns
    class_lst2 = [col_names[0]]
    if class_lst is not None:
        class_lst2 = class_lst + class_lst2
    for val in col_names:
        if val in class_lst2:
            data_df[val] = data_df[val].astype('str')
        else:
            data_df[val] = data_df[val].astype('float')
    id_keep = []  # keep the id both in the agmat file and pheno file, without missing phenotypes and covariates.
    for i in range(data_df.shape[0]):
        if data_df.iloc[i, 0] in id_in_agmat:
            id_keep.append(True)
        else:
            id_keep.append(False)
    data_df = data_df.iloc[id_keep, :]
    id_keep = list(data_df.iloc[:, 0])
    y = np.array(data_df.iloc[:, -1]).reshape(-1, 1)
    bye_mat = None
    xmat = np.ones((y.shape[0], 1))
    if bye is not None:
        xmat = dmatrix(bye, data_df)
        xmat = np.asarray(xmat).reshape(y.shape[0], -1)
        bye_mat = xmat[:, 1:]
    if fix is not None:
        xmat = dmatrix(fix, data_df)
        xmat = np.asarray(xmat)
        _, r = np.linalg.qr(xmat)
        indexes = np.absolute(np.diag(r)) >= 1e-10
        xmat = xmat[:, indexes]
    val = 0
    id_dct = {}
    col = []
    for v in id_keep:
        if v not in id_dct:
            id_dct[v] = val
            val += 1
        col.append(id_dct[v])
    row = range(len(id_keep))
    zmat = csc_matrix(([1.0]*len(row), (row, col)))
    ag = np.zeros((len(id_dct), len(id_dct)))
    with open(agmat_file + '.agrm.id_fmt') as fin:
        for line in fin:
            arr = line.split()
            try:
                ag[id_dct[arr[0]], id_dct[arr[1]]] = ag[id_dct[arr[1]], id_dct[arr[0]]] = float(arr[2])
            except Exception as e:
                del e
    return y, xmat, zmat, ag, id_keep, bye_mat


def _weight_emai_vmat(y, xmat, zmat, ag, init=None, maxiter=100, cc_par=1.0e-8):
    """
    Estimate variance parameters with vmat based on em
    :param y: Phenotypic vector
    :param xmat: the design matrix for fixed effect
    :param zmat: Design matrix for random effects
    :param ag: Genomic relationship matrix
    :param init: Default is None. A list for initial values of variance components
    :param maxiter: Default is 100. The maximum number of iteration times.
    :param cc_par: Convergence criteria for update vector.
    :return: The estimated variances.
    """
    logging.info("#####Prepare#####")
    var_com = np.array(init)
    num_var = len(var_com)
    zgzmat_lst = [zmat.dot((zmat.dot(ag)).T), zmat.dot(zmat.T)]
    logging.info('Initial variances: {}'.format(var_com))
    logging.info("#####Start the iteration#####")
    weight_vec = list(np.arange(0, 1, 0.01)) + [1.0]
    iter = 0
    cc_par_val = 1000.0
    delta = 1000.0
    while iter < maxiter:
        iter += 1
        logging.info('\n\nRound: {:d}'.format(iter))
        # V, the inverse of V and P
        vmat = zgzmat_lst[0] * var_com[0] + zgzmat_lst[1].toarray() * var_com[1] + np.diag([var_com[-1]] * y.shape[0])
        vmat = linalg.inv(vmat)
        vxmat = np.dot(vmat, xmat)
        xvxmat = np.dot(xmat.T, vxmat)
        xvxmat = linalg.inv(xvxmat)
        pmat = vmat - reduce(np.dot, [vxmat, xvxmat, vxmat.T])
        pymat = np.dot(pmat, y)
        # first partial derivatives and working variates
        fd_mat = np.zeros(num_var)
        wv_mat = []
        fd_mat[-1] = np.trace(pmat) - np.sum(np.dot(pymat.T, pymat))
        fd_mat[-1] = -0.5 * fd_mat[-1]
        fd_mat[0] = np.sum(pmat * zgzmat_lst[0]) - np.sum(reduce(np.dot, [pymat.T, zgzmat_lst[0], pymat]))
        fd_mat[0] = -0.5 * fd_mat[0]
        wv_mat.append(np.dot(zgzmat_lst[0], pymat))
        fd_mat[1] = np.sum(zgzmat_lst[1].multiply(pmat)) - np.sum(np.dot(pymat.T, zgzmat_lst[1].dot(pymat)))
        fd_mat[1] = -0.5 * fd_mat[1]
        wv_mat.append(zgzmat_lst[1].dot(pymat))
        wv_mat.append(pymat)
        wv_mat = np.concatenate(wv_mat, axis=1)
        ai_mat = 0.5 * reduce(np.dot, [wv_mat.T, pmat, wv_mat])
        em_mat = []
        for k in range(num_var - 1):
            em_mat.append(ag.shape[0] / (2 * var_com[k] * var_com[k]))
        em_mat.append(y.shape[0] / (2 * var_com[-1] * var_com[-1]))
        em_mat = np.diag(em_mat)
        for weight in weight_vec:
            wemai_mat = weight * em_mat + (1 - weight) * ai_mat
            delta = np.dot(linalg.inv(wemai_mat), fd_mat)
            var_com_new = var_com + delta
            if min(var_com_new) > 0:
                logging.info('EM weight value: ' + str(weight))
                break
        var_com_new = var_com + delta
        logging.info("Updated variances: {}".format(var_com_new))
        cc_par_val = np.sum(delta * delta) / np.sum(np.array(var_com_new) * np.array(var_com_new))
        cc_par_val = np.sqrt(cc_par_val)
        var_com = np.array(var_com_new)
        if cc_par_val < cc_par:
            break
    if cc_par_val < cc_par:
        logging.info('Variances converged.')
        convergence=True
    else:
        logging.info('Variances not converged.')
        convergence=False
    vmat = zgzmat_lst[0] * var_com[0] + zgzmat_lst[1].toarray() * var_com[1] + np.diag([var_com[-1]] * y.shape[0])
    return var_com, convergence, vmat



import sys
import pandas as pd
import numpy as np
from scipy import linalg
import logging

from ._common import _design_matrix, _leg_mt, _full_col_rank

def _pre_data(data_file, id, tpoint, trait, gmat_file, na_method='omit'):
    logging.info('Data file: ' + data_file)
    data_df = pd.read_csv(data_file, sep='\s+', header=0)
    logging.info('NA method: ' + na_method)
    if na_method == 'omit':
        data_df = data_df.dropna()
    elif na_method == 'include':
        data_df = data_df.fillna(method='ffill')
        data_df = data_df.fillna(method='bfill')
    else:
        logging.info('na_method does not exist: ' + na_method)
        sys.exit()
    col_names = data_df.columns
    logging.info('The column names of data file: {}'.format(col_names))
    logging.info('Note: Variates beginning with a capital letter is converted into factors.')
    class_vec = []
    for val in col_names:
        if not val[0].isalpha():
            logging.info("The first character of columns names must be alphabet!")
            sys.exit()
        if val[0] == val.capitalize()[0]:
            class_vec.append(val)
            data_df[val] = data_df[val].astype('str')
        else:
            try:
                data_df[val] = data_df[val].astype('float')
            except Exception as e:
                logging.info(e)
                logging.info("{} may contain string, please check!".format(val))
                sys.exit()
    data_df = data_df.sort_values(by=[id, tpoint])  # sort the data frame by id and time points.
    id_in_gmat = list(pd.read_csv(gmat_file + '.id', sep='\s+', header=None, dtype={0: 'str'}).iloc[:, 0])
    logging.info('Individual column: ' + id)
    if id not in class_vec:
        logging.info('The initial letter of {} should be capital'.format(id))
        sys.exit()
    id_order = []   # id in order
    ind_keep = []
    for val in data_df[id]:
        if val in id_in_gmat:
            ind_keep.append(True)
            if val not in id_order:
                id_order.append(val)
        else:
            ind_keep.append(False)
    data_df = data_df.iloc[ind_keep, :]
    logging.info('Time points column: ' + tpoint)
    if tpoint in class_vec:
        logging.info('The initial letter of {} should be lowercase'.format(tpoint))
        sys.exit()
    logging.info('Trait column: ' + trait)
    if trait in class_vec:
        logging.info('The initial letter of {} should be lowercase'.format(trait))
        sys.exit()
    logging.info("The genomic relationship matrix file is: {}.agrm.id_fmt".format(gmat_file))
    id_dct = dict(zip(id_order, range(len(id_order))))
    ag = np.zeros((len(id_order), len(id_order)))
    with open(gmat_file + '.agrm.id_fmt') as fin:
        for line in fin:
            arr = line.split()
            try:
                ag[id_dct[arr[0]], id_dct[arr[1]]] = ag[id_dct[arr[1]], id_dct[arr[0]]] = float(arr[2])
            except Exception as e:
                del e
    id_code = []  # code the id in the data file from 0 to ...
    for val in data_df[id]:
        id_code.append(id_dct[val])
    return data_df, ag, id_code, id_order



import logging
from gmat.gmatrix import agmat
from gmat.uvlmm.gwas import fixed, fixed_gbye, fixed_, fixed_repeat
logging.basicConfig(level=logging.INFO)
bed_file = 'CobaltChloride'
pheno_file = 'CobaltChloride3'
# id-id-value form
agmat_file = 'test'

agmat2 = agmat(bed_file, out_file=agmat_file, inv=False, small_val=0.001, out_fmt='row_col_val')
"""
# 和GEMMA相同
fixed(pheno_file, bed_file, agmat_file, class_lst=['batch'], fix="batch",
out_file='res1.txt', condition_snp=None, npart=100, maxiter0=100, maxiter1=10,
          speed=True, pcut=5.0e-5)

# conditional GWAS
fixed(pheno_file, bed_file, agmat_file, class_lst=None, fix="batch",
out_file='res2.txt', condition_snp=None, npart=100, maxiter0=100, maxiter1=10,
          speed=True, pcut=5.0e-5)

# 基因环境互作、不加永久环境效应，batch是连续变量
bye = 'batch'
fixed_gbye(pheno_file, bed_file, agmat_file, bye, class_lst=None, fix=None, out_file='res3.txt', condition_snp=None, npart=100,
          maxiter0=100, maxiter1=10, speed=True, pcut=5.0e-5)
"""
# 基因环境互作、不加永久环境效应，batch是分类变量
bye = 'batch'
fixed_gbye(pheno_file, bed_file, agmat_file, bye, class_lst=["batch"], fix=None, out_file='res4.txt', condition_snp=None, npart=100,
          maxiter0=100, maxiter1=10, speed=True, pcut=5.0e-5)


"""
# 基因环境互作、加永久环境效应，batch是连续变量
fixed_repeat(pheno_file, bed_file, agmat_file, bye=bye, class_lst=None, fix=None, out_file='res5.txt', condition_snp=None, npart=100,
          maxiter=100)


# 基因环境互作、加永久环境效应，batch是分类变量
fixed_repeat(pheno_file, bed_file, agmat_file, bye=bye, class_lst=["batch"], fix=None, out_file='res6.txt', condition_snp=None, npart=100,
          maxiter=100)
"""



import sys
import math
import numpy as np
import pandas as pd
from scipy import linalg
from scipy import sparse
from scipy.sparse import csr_matrix, block_diag, bmat
import logging
from functools import reduce


def _emai(y_lst, xmat_lst, zmat_lst, ag, error_lst, init=None, maxiter=200, cc_par=1.0e-8,
                      cc_gra=1.0e-3, em_weight_step=0.01, out_file='var_com'):
    def init_var(y_lst):
        """
        initialize the variances
        """
        y_var_lst = np.array(list(map(np.var, y_lst)))/2
        add_cov = np.diag(y_var_lst)
        add_ind = np.tril_indices_from(add_cov)
        error_cov = np.diag(y_var_lst)
        error_ind = np.tril_indices_from(error_cov)
        cov_ind = [1] * len(add_ind[0]) + [2] * len(error_ind[0])
        cov_ind_i = list(add_ind[0] + 1) + list(error_ind[0] + 1)
        cov_ind_j = list(add_ind[1] + 1) + list(error_ind[1] + 1)
        var_com = list(add_cov[add_ind]) + list(error_cov[error_ind])
        var_df = {'var': cov_ind,
                  "vari": cov_ind_i,
                  "varj": cov_ind_j,
                  "var_com": var_com}
        var_df = pd.DataFrame(var_df, columns=['var', "vari", "varj", "var_com"])
        return np.array(var_com), var_df
    var_com, var_df = init_var(y_lst)
    if init is not None:
        if len(var_com) != len(init):
            logging.error('The length of initial variances should be {}'.format(len(var_com)))
            sys.exit()
        else:
            var_com = np.array(init)
            var_df['var_com'] = var_com

    def pre_cov(y_lst, var_df):
        """
        return the covariance matrix. If one covariance matrix is not positive, return None.
        """
        add_cov = np.zeros((len(y_lst), len(y_lst)))
        add_cov[np.tril_indices(len(y_lst))] = np.array(var_df[var_df['var']==1].iloc[:, -1])
        add_cov = np.add(add_cov, np.tril(add_cov, -1).T)
        error_cov = np.zeros((len(y_lst), len(y_lst)))
        error_cov[np.tril_indices(len(y_lst))] = np.array(var_df[var_df['var'] == 2].iloc[:, -1])
        error_cov = np.add(error_cov, np.tril(error_cov, -1).T)
        try:
            linalg.cholesky(add_cov)
            linalg.cholesky(error_cov)
        except Exception as _:
                return None
        return add_cov, error_cov

    cov_lst = pre_cov(y_lst, var_df)
    if cov_lst is None:
        logging.error("ERROR: Initial variances is not positive define, please check!")
        sys.exit()
    add_cov, error_cov = cov_lst
    xmat_con = np.array(xmat_lst[0])
    for i in range(1, len(xmat_lst)):
        xmat_con = linalg.block_diag(xmat_con, xmat_lst[i])
    zmat_con = block_diag(zmat_lst, format='csr')
    y_con = np.concatenate(y_lst, axis=0)
    iter_count = 0
    cc_par_val = 1000.0
    cc_gra_val = 1000.0
    delta = 1000.0
    var_com_update = var_com * 1000
    logging.info("initial variances: {}".format(var_df))
    while iter_count < maxiter:
        iter_count += 1
        logging.info('***Start the iteration: {} ***'.format(iter_count))
        logging.info("Calculate V and P matrix")
        vmat = []
        for i in range(len(error_lst)):
            vec = []
            for j in range(len(error_lst[i])):
                vec.append(error_lst[i][j]*error_cov[i, j])
            vmat.append(vec)
        vmat = bmat(vmat, format='csr')
        vmat += zmat_con.dot((zmat_con.dot(linalg.kron(add_cov, ag))).T)
        vmat = linalg.inv(vmat)
        vxmat = np.dot(vmat, xmat_con)
        xvxmat = np.dot(xmat_con.T, vxmat)
        xvxmat = linalg.inv(xvxmat)
        pmat = vmat - reduce(np.dot, [vxmat, xvxmat, vxmat.T])
        pymat = np.dot(pmat, y_con)
        fd_mat = []
        wv_mat = []
        for i in range(add_cov.shape[0]):
            for j in range(i+1):
                add_cov_zeros = np.zeros((add_cov.shape[0], add_cov.shape[1]))
                add_cov_zeros[i, j] = add_cov_zeros[j, i] = 1.0
                temp = zmat_con.dot((zmat_con.dot(linalg.kron(add_cov_zeros, ag))).T)
                wv_mat.append(np.dot(temp, pymat))
                fd_val = -np.sum(pmat * temp) + np.sum(np.dot(pymat.T, wv_mat[-1]))
                fd_mat.append(fd_val*0.5)
        for i in range(error_cov.shape[0]):
            for j in range(i+1):
                error_cov_zeros = np.zeros((error_cov.shape[0], error_cov.shape[1]))
                error_cov_zeros[i, j] = error_cov_zeros[j, i] = 1.0
                error_mat = []
                for m in range(len(error_lst)):
                    vec = []
                    for n in range(len(error_lst[m])):
                        vec.append(error_lst[m][n] * error_cov_zeros[m, n])
                    error_mat.append(vec)
                error_mat = bmat(error_mat, format='csr')
                wv_mat.append(error_mat.dot(pymat))
                fd_val = -np.sum(pmat * error_mat.toarray()) + np.sum(np.dot(pymat.T, wv_mat[-1]))
                fd_mat.append(fd_val * 0.5)
        fd_mat = np.array(fd_mat)
        np.savetxt(out_file + '.fd', fd_mat)
        wv_mat = np.concatenate(wv_mat, axis=1)
        ai_mat = reduce(np.dot, [wv_mat.T, pmat, wv_mat])
        np.savetxt(out_file + '.ai', ai_mat)
        ind = np.tril_indices_from(add_cov)
        len_add_cov = len(ind[0])
        em_mat_add = np.zeros((len_add_cov, len_add_cov))
        for i in range(len_add_cov):
            for j in range(i+1):
                em_mat_add[i, j] = (add_cov[ind[0][i], ind[0][j]] * add_cov[ind[1][i], ind[1][j]] +
                                  add_cov[ind[0][i], ind[1][j]] * add_cov[ind[1][i], ind[0][j]]) / (2.0 * ag.shape[0])
        ind = np.tril_indices_from(error_cov)
        len_error_cov = len(ind[0])
        em_mat_error = np.zeros((len_error_cov, len_error_cov))
        for i in range(len_error_cov):
            for j in range(i + 1):
                em_mat_error[i, j] = (error_cov[ind[0][i], ind[0][j]] * error_cov[ind[1][i], ind[1][j]] +
                                    error_cov[ind[0][i], ind[1][j]] * error_cov[ind[1][i], ind[0][j]]) \
                        / (len(error_lst[ind[0][i]][ind[0][j]].data) + len(error_lst[ind[1][i]][ind[1][j]].data))
        em_mat = linalg.block_diag(em_mat_add, em_mat_error)
        em_mat = 2.0 * em_mat
        em_mat += np.tril(em_mat, k=-1).T
        em_mat = linalg.inv(em_mat)
        np.savetxt(out_file + '.em', em_mat)
        # Increase em weight to guarantee variances positive
        gamma = -em_weight_step
        while gamma < 1.0:
            gamma = gamma + em_weight_step
            if gamma >= 1.0:
                gamma = 1.0
            wemai_mat = (1 - gamma) * ai_mat + gamma * em_mat
            delta = np.dot(linalg.inv(wemai_mat), fd_mat)
            var_com_update = var_com + delta
            var_df['var_com'] = var_com_update
            cov_lst = pre_cov(y_lst, var_df)
            if cov_lst is not None:
                logging.info('EM weight value: ' + str(gamma))
                break
        var_df.to_csv(out_file + '.var', sep='\t', index=False)
        if cov_lst is None:
            logging.error("ERROR: Updated variances is not positive define!")
            sys.exit()
        add_cov, error_cov = cov_lst
        # Convergence criteria
        cc_par_val = np.sum(pow(delta, 2)) / np.sum(pow(var_com_update, 2))
        cc_par_val = np.sqrt(cc_par_val)
        cc_gra_val = np.sqrt(np.sum(pow(fd_mat, 2))) / len(var_com)
        var_com = var_com_update.copy()
        logging.info("Change in parameters: {}, Norm of gradient vector: {}".format(cc_par_val, cc_gra_val))
        if cc_par_val < cc_par and cc_gra_val < cc_gra:
            break
    if cc_par_val < cc_par and cc_gra_val < cc_gra:
        logging.info("Variances Converged")
    else:
        logging.info("Variances not Converged")
    var_df['var_com'] = var_com
    return var_df




### Estimate the variances
import numpy as np
import pandas as pd
from gmat.gmatrix import agmat
from gmat.longwas.balance import balance_varcom
import logging

logging.basicConfig(level=logging.INFO)

bed_file = '../data/mouse_long/plink'
out_file = '../data/mouse_long/test'
#agmat(bed_file, out_file=out_file, inv=False, small_val=0.001, out_fmt='id_id_val')

data_file = '../data/mouse_long/phe.balance.txt'
id = 'ID'
tpoint = np.array(range(16)) + 1.0
trait = np.array(np.arange(1, 17), dtype='str')
trait = list(map(lambda x, y: x + y, ['trait']*16, trait))
agmat_file = '../data/mouse_long/test'
prefix_outfile = '../data/mouse_long/balance_varcom'
res_var = balance_varcom(data_file, id, tpoint, trait, agmat_file, tfix=None, prefix_outfile=prefix_outfile)
res_var = balance_varcom(data_file, id, tpoint, trait, agmat_file, tfix="Sex", prefix_outfile=prefix_outfile+".tSex")
res_var = balance_varcom(data_file, id, tpoint, trait, agmat_file, fix="Sex", prefix_outfile=prefix_outfile+".Sex")
print(res_var)

### longitudinal GWAS by fixed method

from gmat.longwas.balance import balance_longwas_fixed
var_com = pd.read_csv('../data/mouse_long/balance_varcom.var', header=0, sep='\s+')
prefix_outfile = '../data/mouse_long/balance_longwas_fixed'
res_fixed = balance_longwas_fixed(data_file, id, tpoint, trait, kin_file, bed_file, var_com, snp_lst=None,
                                 prefix_outfile=prefix_outfile)

### longitudinal GWAS by trans method

from gmat.longwas.balance import balance_longwas_trans
var_com = pd.read_csv('../data/mouse_long/balance_varcom.var', header=0, sep='\s+')
prefix_outfile = '../data/mouse_long/balance_longwas_trans'
res_trans = balance_longwas_trans(data_file, id, tpoint, trait, kin_file, bed_file, var_com, snp_lst=None,
                                 prefix_outfile=prefix_outfile)



import sys
import math
import numpy as np
import pandas as pd
from scipy import linalg
from scipy import sparse
from scipy.stats import chi2
from scipy.sparse import csr_matrix, block_diag, bmat
import logging
from tqdm import tqdm
from functools import reduce
from gmat.process_plink.process_plink import impute_geno
from pysnptools.snpreader import Bed


from gmat.mvlmm.unbalance._pre_data import _pre_data


def fixed(data_file, id, trait, bed_file, agmat_file, var_com, snp_lst=None, fix=None, prefix_outfile='fixed'):
    """
    Estimate the variances for the multivariate linear mixed model.
    :param data_file:
    :param trait:
    :param agmat_file:
    :param fix:
    :param init:
    :param maxiter:
    :param cc_par:
    :param cc_gra:
    :param em_weight_step:
    :param prefix_outfile:
    :return:
    """
    logging.info("######################")
    logging.info("###Prepare the data###")
    logging.info("######################")
    y_lst, xmat_lst, zmat_lst, ag, error_lst, id_raw_lst = _pre_data(data_file, id, trait, agmat_file, fix=fix)
    def pre_cov(y_lst, var_df):
        """
        return the covariance matrix. If one covariance matrix is not positive, return None.
        """
        add_cov = np.zeros((len(y_lst), len(y_lst)))
        add_cov[np.tril_indices(len(y_lst))] = np.array(var_df[var_df['var']==1].iloc[:, -1])
        add_cov = np.add(add_cov, np.tril(add_cov, -1).T)
        error_cov = np.zeros((len(y_lst), len(y_lst)))
        error_cov[np.tril_indices(len(y_lst))] = np.array(var_df[var_df['var'] == 2].iloc[:, -1])
        error_cov = np.add(error_cov, np.tril(error_cov, -1).T)
        try:
            linalg.cholesky(add_cov)
            linalg.cholesky(error_cov)
        except Exception as _:
                return None
        return add_cov, error_cov
    add_cov, error_cov = pre_cov(y_lst, var_com)
    xmat_con = np.array(xmat_lst[0])
    for i in range(1, len(xmat_lst)):
        xmat_con = linalg.block_diag(xmat_con, xmat_lst[i])
    zmat_con = block_diag(zmat_lst, format='csr')
    y_con = np.concatenate(y_lst, axis=0)
    vmat = []
    for i in range(len(error_lst)):
        vec = []
        for j in range(len(error_lst[i])):
            vec.append(error_lst[i][j] * error_cov[i, j])
        vmat.append(vec)
    vmat = bmat(vmat, format='csr')
    vmat += zmat_con.dot((zmat_con.dot(linalg.kron(add_cov, ag))).T)
    vmat = linalg.inv(vmat)
    vxmat = np.dot(vmat, xmat_con)
    xvxmat = np.dot(xmat_con.T, vxmat)
    xvxmat = linalg.inv(xvxmat)
    pmat = vmat - reduce(np.dot, [vxmat, xvxmat, vxmat.T])
    snp_on_disk = Bed(bed_file, count_A1=False)
    fam_df = pd.read_csv(bed_file + '.fam', sep='\s+', header=None)
    id_geno = list(np.array(fam_df.iloc[:, 1], dtype=str))
    id_order_index = []
    for i in id_raw_lst:
        id_order_index.append(id_geno.index(i))
    logging.info('***Read the snp data***')
    num_id = snp_on_disk.iid_count
    num_snp = snp_on_disk.sid_count
    logging.info("There are {:d} individuals and {:d} SNPs.".format(num_id, num_snp))
    if snp_lst is None:
        snp_lst = range(num_snp)
    snp_lst = list(snp_lst)
    if min(snp_lst) < 0 or max(snp_lst) >= num_snp:
        logging.info('The value in the snp list should be >= {} and < {}', 0, num_snp)
        exit()
    snp_mat = snp_on_disk[:, snp_lst].read().val
    if np.any(np.isnan(snp_mat)):
        logging.info('Missing genotypes are imputed with random genotypes.')
        snp_mat = impute_geno(snp_mat)
    bye_mat = np.zeros((len(id_raw_lst), len(y_lst)-1))
    indexa = len(y_lst[0])
    for i in range(1, len(y_lst)):
        indexb = indexa + len(y_lst[i])
        bye_mat[indexa:indexb, i-1] = 1.0
        indexa = indexb
    logging.info('########################################################')
    logging.info('###Start the fixed regression GWAS for unbalance data###')
    logging.info('########################################################')
    chi_df = len(y_lst)
    eff_main_vec = []
    se_main_vec = []
    p_main_vec = []
    eff_vec = []
    p_gbye_vec = []
    p_vec = []
    for i in tqdm(range(snp_mat.shape[1])):
        snp_fix = snp_mat[id_order_index, i:(i+1)]
        p_xsnp = np.dot(pmat, snp_fix)
        xpx = np.dot(snp_fix.T, p_xsnp)
        se_main = np.sqrt(xpx[0, 0])
        xpy = np.dot(p_xsnp.T, y_con)
        eff_main = xpy[-1, -1]
        chi_val = eff_main*eff_main/(se_main*se_main)
        p_main = chi2.sf(chi_val, 1)
        eff_main_vec.append(eff_main)
        se_main_vec.append(se_main)
        p_main_vec.append(p_main)
        snp_fix = np.concatenate([snp_fix, snp_fix*bye_mat], axis=1)
        p_xsnp = np.dot(pmat, snp_fix)
        xpx = np.dot(snp_fix.T, p_xsnp)
        xpx = linalg.inv(xpx)
        xpy = np.dot(p_xsnp.T, y_con)
        b = np.dot(xpx, xpy)
        eff = b[-(chi_df-1):, -1]
        eff_var = xpx[-(chi_df-1):, -(chi_df-1):]
        chi_val = np.sum(np.dot(np.dot(eff.T, linalg.inv(eff_var)), eff))
        p_gbye = chi2.sf(chi_val, chi_df-1)
        p_gbye_vec.append(p_gbye)
        eff = b[-chi_df:, -1]
        eff_var = xpx[-chi_df:, -chi_df:]
        chi_val = np.sum(np.dot(np.dot(eff.T, linalg.inv(eff_var)), eff))
        p_val = chi2.sf(chi_val, chi_df)
        eff_vec.append(eff)
        p_vec.append(p_val)
    logging.info('Finish association analysis')
    logging.info('***Output***')
    snp_info_file = bed_file + '.bim'
    snp_info = pd.read_csv(snp_info_file, sep='\s+', header=None)
    res_df = snp_info.iloc[snp_lst, [0, 1, 3, 4, 5]]
    res_df.columns = ['chro', 'snp_ID', 'pos', 'allele1', 'allele2']
    res_df.loc[:, 'order'] = snp_lst
    res_df = res_df.iloc[:, [5, 0, 1, 2, 3, 4]]
    res_df.loc[:, 'eff_main'] = eff_main_vec
    res_df.loc[:, 'se_main'] = se_main_vec
    res_df.loc[:, 'p_main'] = p_main_vec
    eff_vec = np.array(eff_vec)
    for i in range(eff_vec.shape[1]):
        col_ind = 'eff' + str(i)
        res_df.loc[:, col_ind] = eff_vec[:, i]
    res_df.loc[:, 'p_gbye'] = p_gbye_vec
    res_df.loc[:, 'p_val'] = p_vec
    out_file = prefix_outfile + '.res'
    res_df.to_csv(out_file, sep='\t', index=False)
    return res_df













import numpy as np
from scipy import linalg
from functools import reduce
import logging
from scipy.sparse import hstack


def fisher_vmat(y, xmat, zmat_lst, gmat_lst, init=None, maxiter=100, cc_par=1.0e-8):
    """
    :param y:
    :param xmat:
    :param zmat_lst:
    :param gmat_lst:
    :param init:
    :param maxiter:
    :param cc_par:
    :return:
    """
    logging.info("#####Prepare#####")
    num_var = len(gmat_lst) + 1
    var_com = np.ones(num_var)
    if init is not None:
        var_com = np.array(init)
    y = np.array(y).reshape(-1, 1)
    n = y.shape[0]
    xmat = np.array(xmat).reshape(n, -1)
    zgzmat_lst = []
    for val in range(len(gmat_lst)):
        zgzmat_lst.append(zmat_lst[val].dot((zmat_lst[val].dot(gmat_lst[val])).T))
    logging.info('Initial variances: ' + ' '.join(list(np.array(var_com, dtype=str))))
    logging.info("#####Start the iteration#####\n\n")
    iter = 0
    cc_par_val = 1000.0
    while iter < maxiter:
        iter += 1
        logging.info('Round: {:d}'.format(iter))
        # V, the inverse of V and P
        vmat = np.diag([var_com[-1]] * n)
        for val in range(len(zgzmat_lst)):
            vmat += zgzmat_lst[val] * var_com[val]
        vmat = np.linalg.inv(vmat)
        vxmat = np.dot(vmat, xmat)
        xvxmat = np.dot(xmat.T, vxmat)
        xvxmat = linalg.inv(xvxmat)
        pmat = vmat - reduce(np.dot, [vxmat, xvxmat, vxmat.T])
        pymat = np.dot(pmat, y)
        # first partial derivatives and working variates
        fd_mat = np.zeros(num_var)
        pzgzmat_lst = []
        fd_mat[-1] = np.trace(pmat) - np.sum(np.dot(pymat.T, pymat))
        fd_mat[-1] = -0.5 * fd_mat[-1]
        for i in range(num_var - 1):
            fd_mat[i] = np.sum(pmat*zgzmat_lst[i]) - np.sum(reduce(np.dot, [pymat.T, zgzmat_lst[i], pymat]))
            fd_mat[i] = -0.5 * fd_mat[i]
            pzgzmat_lst.append(np.dot(pmat, zgzmat_lst[i]))
        pzgzmat_lst.append(pmat)
        fisher_mat = np.zeros((num_var, num_var))
        for i in range(len(pzgzmat_lst)):
            for j in range(i + 1):
                fisher_mat[i, j] = -0.5*np.sum(pzgzmat_lst[i] * pzgzmat_lst[j].T)
                fisher_mat[j, i] = fisher_mat[i, j]
        print(fisher_mat)
        delta = np.dot(linalg.inv(fisher_mat), fd_mat)
        var_com_new = var_com - delta
        cc_par_val = np.sum(delta * delta) / np.sum(np.array(var_com_new) * np.array(var_com_new))
        cc_par_val = np.sqrt(cc_par_val)
        var_com = np.array(var_com_new)
        logging.info('Norm of update vector: {:e}'.format(cc_par_val))
        var_com_str = ' '.join(np.array(var_com, dtype=str))
        logging.info('Updated variances: {}'.format(var_com_str))
        if cc_par_val < cc_par:
            break
    if cc_par_val < cc_par:
        logging.info('Variances converged.')
    else:
        logging.info('Variances not converged.')
    return var_com



import numpy as np
import pandas as pd
from gmat.gmatrix import agmat
from gmat.longwas.unbalance import unbalance_varcom
import logging

logging.basicConfig(level=logging.INFO)

bed_file = '../data/mouse_long/plink'
agmat(bed_file, out_file='../data/mouse_long/test', out_fmt='id_id_val')
data_file = '../data/mouse_long/phe.unbalance.txt'
id = 'ID'
tpoint = 'weak'
trait = 'trait'
agmat_file = '../data/mouse_long/test'
tfix = 'Sex'
prefix_outfile = '../data/mouse_long/unbalance_varcom'
res_var = unbalance_varcom(data_file, id, tpoint, trait, agmat_file, tfix=None, out_file=prefix_outfile)
res_var = unbalance_varcom(data_file, id, tpoint, trait, agmat_file, tfix="Sex", out_file=prefix_outfile + '.tSex')
res_var = unbalance_varcom(data_file, id, tpoint, trait, agmat_file, fix="Sex", out_file=prefix_outfile + '.Sex')

#print(res_var)


var_com = pd.read_csv("../data/mouse_long/unbalance_varcom.var", sep='\s+', header=0)

"""
from gmat.longwas.unbalance import unbalance_longwas_trans
prefix_outfile = '../data/mouse_long/unbalance_longwas_trans'
res_trans = unbalance_longwas_trans(data_file, id, tpoint, trait, bed_file, agmat_file, var_com, tfix="Sex",
                                    prefix_outfile=prefix_outfile)


from gmat.longwas.unbalance import unbalance_longwas_trans
prefix_outfile = '../data/mouse_long/unbalance_condition_longwas_trans'
res_trans = unbalance_longwas_trans(data_file, id, tpoint, trait, bed_file, agmat_file, var_com, condition_snp="UNC18848064", tfix="Sex",
                                    prefix_outfile=prefix_outfile)
"""





from gmat.longwas.unbalance import unbalance_longwas_fixed
prefix_outfile = '../data/mouse_long/unbalance_longwas_fixed'
res_fixed = unbalance_longwas_fixed(data_file, id, tpoint, trait, bed_file, agmat_file, var_com,
 tfix="Sex", forder=3, aorder=3, porder=3, na_method='omit', prefix_outfile=prefix_outfile)


from gmat.longwas.unbalance import unbalance_longwas_fixed
prefix_outfile = '../data/mouse_long/unbalance_condition_longwas_fixed'
res_fixed = unbalance_longwas_fixed(data_file, id, tpoint, trait, bed_file, agmat_file, var_com, condition_snp="UNC18848064",
 tfix="Sex", forder=3, aorder=3, porder=3, na_method='omit', prefix_outfile=prefix_outfile)



"""
from gmat.longwas.unbalance import unbalance_longwas_trans
prefix_outfile = '../data/mouse_long/unbalance_longwas_trans'
res_trans = unbalance_longwas_trans(data_file, id, tpoint, trait, bed_file, agmat_file, var_com, tfix="Sex",
                                    prefix_outfile=prefix_outfile)

"""

'''
from gmat.longwas.unbalance import unbalance_condition_longwas_trans

prefix_outfile = '../data/mouse_long/unbalance_condition_longwas_trans8'
res_trans = unbalance_condition_longwas_trans(data_file, id, tpoint, trait, bed_file, agmat_file, 
      var_com, condition_snp="UNC30664376", tfix='Sex', fix="Sex", crank=True,
                                    prefix_outfile=prefix_outfile)
'''
"""
from gmat.longwas.unbalance import unbalance_longwas_fixed_permutation
prefix_outfile = '../data/mouse_long/unbalance_longwas_fixed_permutation'
unbalance_longwas_fixed_permutation(data_file, id, tpoint, trait, bed_file, agmat_file, var_com, permutation_lst=range(10),
            snp_lst=None, tfix="Sex", fix=None, forder=3, aorder=3, porder=3, na_method='omit',
                             prefix_outfile=prefix_outfile)


from gmat.longwas.unbalance import unbalance_longwas_trans_permutation
prefix_outfile = '../data/mouse_long/unbalance_longwas_trans_permutation'
unbalance_longwas_trans_permutation(data_file, id, tpoint, trait, bed_file, agmat_file, var_com, permutation_lst=range(10), snp_lst=None,
            tfix="Sex", fix=None, forder=3, aorder=3, porder=3, na_method='omit',
                             prefix_outfile=prefix_outfile)
"""
